/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/

#include <libc.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/resource.h>

#include <oo.h>
#include <ooIndex.h>
#include "OO7.h"
#include "GenParams.h"
#include "VarParams.h"


//////////////////////////////////////////////////////////////////
//
// Global Variables, etc., for Benchmark Database Generation
//
//////////////////////////////////////////////////////////////////

struct rusage startUsage, endUsage;
struct timeval startWallTime;
struct timeval endWallTime;
struct timezone ignoreTimeZone;


extern void SetParams(char* configFileName);

extern double ComputeWallClockTime(struct timeval *startWallTime, 
	           struct timeval *endWallTime);

extern double ComputeUserTime(struct rusage *startUsage, 
	           struct rusage *endUsage);

extern double ComputeSystemTime(struct rusage *startUsage, 
	           struct rusage *endUsage);

extern void CreateDummy(int cnt);

ooTrans	xact;
extern ooHandle(ooContObj) DocContainerH;


int	nextAtomicId = 1;
int	nextCompositeId = 1;
int	nextComplexAssemblyId = 1;
int	nextBaseAssemblyId = 1;
int	nextModuleId = 1;
int	debugMode = FALSE;

char    *types[NumTypes] = {
		"type000", "type001", "type002", "type003", "type004",
		"type005", "type006", "type007", "type008", "type009"
	};

// data structures to keep track of which composite parts each base
// assemblies uses.  This is done so that we can generate the base
// assemblies first and then the composite parts.  Why bother?
// In particular, if you did the composite parts first and the base
// assemblies, there would be no need for these data structures.
// This second approach worked fine for the small and medium databases
// but not for the large as the program seeked all over the disk.

BAIdList* private_cp;
BAIdList* shared_cp;

main(int argc, char** argv)
{
	if (argc < 2) 
	{
	      fprintf(stderr, "Usage: %s <configFileName> [-d]\n", argv[0]);
	      exit(1);
	}
	// See if debug mode is desired
	if ((argc == 3) && (strcmp(argv[2], "-d") == 0)) debugMode = TRUE;
	else  debugMode = FALSE;
	
	SetParams(argv[1]);

	ooInit(12, CachePages, CachePages); // initialize cache size
	ooSetIndexMode(oocExplicitUpdate); // set explicit index updates

	printf("Initialized system with call ooInit(12, %d, %d)\n",
		CachePages, CachePages);

        // get wall clock time
        gettimeofday(&startWallTime, &ignoreTimeZone);

	shared_cp = new BAIdList[TotalCompParts+1];
	private_cp = new BAIdList[TotalCompParts+1];

	//////////////////////////////////////////////////////////////////
	//
	// Initialize Objectivity/DB
	//
	//////////////////////////////////////////////////////////////////

	// Start the transaction
	xact.start();

	// get starting usage values.
	getrusage(RUSAGE_SELF, &startUsage);

	// Open the federated database
	ooHandle(ooFDObj)	federatedDBH;
	if (federatedDBH.open("OO7", oocUpdate) != oocSuccess) {
	   fprintf(stderr, "ERROR: Cannot open OO7 Federated Database.\n");
	   xact.abort();
	   exit(1);
	}

	//////////////////////////////////////////////////////////////////
	//
	// Replace and Open the Database
	//
	//////////////////////////////////////////////////////////////////

	ooHandle(ooDBObj)	databaseH;
	
	databaseH = ooReplace(ooDBObj, ("OO7_DB", 0, 0, "", "/n/sparc6/raw/data_fs"), federatedDBH);
	if (!databaseH) {
	      fprintf(stderr, "ERROR: Cannot open OO7_DB Database.\n");
	      xact.abort();
	      exit(1);
	}



	//////////////////////////////////////////////////////////////////
	//
	// Do Benchmark Data Generation
	//
	//////////////////////////////////////////////////////////////////

	// Generate a container with the desired number of composite parts,
	// first setting up appropriate secondary indices for OO7 queries

	const int clusterFactor = 30;	// Objectivity recommended 30

	int bytesPerComposite = int (2.5 * (sizeof(Document) +
				NumAtomicPerComp * (sizeof(AtomicPart) +
			        (NumConnPerAtomic * sizeof(Connection)))));

	int DesignLibraryPages = 1 +
             (bytesPerComposite*NumCompPerModule*TotalModules)/PageSize;

	printf("Creating DesignLibrary container with %d pages.\n",
	       DesignLibraryPages);

	int DocPages = 1 + (int) (1.1 * (sizeof(Document) * NumCompPerModule * 
		TotalModules) / PageSize);

	ooHandle(ooContObj) contH = new("DesignLibrary", clusterFactor, 
                                         DesignLibraryPages, 0, databaseH)
	        		    ooContObj;

	ooHandle(ooKeyDesc) dateKeyH;
	dateKeyH = new (contH) ooKeyDesc(ooTypeN(AtomicPart), oocFalse);
	ooHandle(ooKeyField) dateFieldH;
	dateFieldH = new (dateKeyH) ooKeyField(ooTypeN(AtomicPart),"buildDate");
	dateKeyH->addField(dateFieldH);
	dateKeyH->createIndex(contH);

	// create separate container for the documents
        printf("Creating Document container with %d pages\n", DocPages);
        DocContainerH = new("DocContainer", clusterFactor, DocPages, 0, databaseH) 
		ooContObj;

	ooHandle(ooKeyDesc) docIdKeyH;
	docIdKeyH = new (DocContainerH) ooKeyDesc(ooTypeN(Document), oocTrue);
	ooHandle(ooKeyField) docIdFieldH;
        docIdFieldH = new (docIdKeyH) ooKeyField(ooTypeN(Document),"id");
        docIdKeyH->addField(docIdFieldH);
	docIdKeyH->createIndex(DocContainerH);

        // set random number generator
        srandom(1);


	// First generate the desired number of modules
	char moduleName[40];
	while (nextModuleId <= TotalModules) {
	    printf("Generating module %d.\n", nextModuleId);
	    int id = nextModuleId++;
            sprintf(moduleName, "Module %08d", id);
	    ooHandle(Module) moduleH = new(moduleName, 1, 25, 0,databaseH)
	        		       Module (id);
	}

        int id;
        ooHandle(CompositePart) compH;

        ooKey key;
        key.type = oocInt32;
        key.offset = ooGetMemberOffset(CompositePart, id);
        key.size   = ooGetMemberSize(CompositePart, id);
        key.value  = &id;


	// next generate all the composite parts
	while (nextCompositeId <= TotalCompParts) 
	{
	    if (nextCompositeId % 500 == 0) {
		printf("Generated %d composite parts\n", nextCompositeId);
            }
	    id = nextCompositeId++;
	    compH = ooNewKey(CompositePart, (id), contH, key);
        }

	// Print out some useful information about what happened

	printf("=== DONE CREATING DATABASE, TOTALS WERE ===\n");
	printf("    # atomic parts\t\t%d\n", nextAtomicId-1);
	printf("    # composite parts\t\t%d\n", nextCompositeId-1);
	printf("    # complex assemblies\t%d\n", nextComplexAssemblyId-1);
	printf("    # base assemblies\t\t%d\n", nextBaseAssemblyId-1);
	printf("    # modules\t\t\t%d\n", nextModuleId-1);

	//////////////////////////////////////////////////////////////////
	//
	// Shutdown Objectivity/DB
	//
	//////////////////////////////////////////////////////////////////

	// Close the Federated Database

	federatedDBH.close();

	// Commit the transaction
	xact.commit();

        // compute and report wall clock time
        gettimeofday(&endWallTime, &ignoreTimeZone);
        printf("Wall-Clock time to generate database: %f seconds.\n", 
	             ComputeWallClockTime(&startWallTime, &endWallTime));

	getrusage(RUSAGE_SELF, &endUsage);
	fprintf(stdout, "(CPU: %f seconds.)\n", 
	                ComputeUserTime(&startUsage, &endUsage) +
			ComputeSystemTime(&startUsage, &endUsage));

//        ooRunStatus();

	// Exit

        // create 60,000 1000 byte dummy objects
	char* purgeVar;
	purgeVar = getenv("PURGE");
	if (purgeVar != NULL) {
	    if (atoi(purgeVar) == 1) {
        	printf("creating 60000 dummy 1,000 byte objects\n");
		CreateDummy (60000);
	    }
	    else printf("not creating dummy purge collection\n");
	}
	else printf("not creating dummy purge collection\n");

	// Exit
	exit(0);
}
