/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/


#include <libc.h>
#include <string.h>
#include <stdio.h>

#include <oo.h>
#include "OO7.h"
#include "GenParams.h"
#include "BenchParams.h"

//////////////////////////////////////////////////////////////////
//
// Query #6 - find all assemblies (base or complex) B that "use" 
// (directly or transitively ) a composite part with a more recent 
// build date than B's build date.
//
/////////////////////////////////////////////////////////////////


extern int	debugMode;


//////////////////////////////////////////////////////////////////
//
// Recursive Workhorse Routine for Query #6
//
//////////////////////////////////////////////////////////////////


int checkMaxCompDate(ooHandle(Assembly)& assmH, int& count)
{
    // check to see if this assembly uses a more recently dated composite
    // part, and process it if so - handling complex and base assemblies
    // differently since the former have subassemblies while the latter
    // directly use composite parts

    int maxSubDate = -1;

    if (assmH.typeN() == ooTypeN(ComplexAssembly)) {

        // complex assembly case - check its subassemblies recursively

        if (debugMode) {
            printf("In Query6:  complex assembly %d, buildDate = %d\n",
                    assmH->id, assmH->buildDate);
        }

	ooHandle(ComplexAssembly) complH = (ooHandle(ComplexAssembly)&) assmH;
        ooItr(Assembly) assmI;
	for (complH->subAssemblies(assmI); assmI.next(); ) {
	    int rsltDate = checkMaxCompDate(assmI, count);
	    if (rsltDate > maxSubDate) {
	        maxSubDate = rsltDate;
	    }
	}

    } else {

        // base assembly case - check its composite part build dates

        if (debugMode) {
            printf("In Query6:  base assembly %d, buildDate = %d\n",
                    assmH->id, assmH->buildDate);
            printf("    composite part dates = { ");
        }

	ooHandle(BaseAssembly) baseH = (ooHandle(BaseAssembly)&) assmH;
        ooItr(CompositePart) compI;
	for (baseH->componentsPriv(compI); compI.next(); ) {
            if (debugMode) { printf("%d ",compI->buildDate); }
	    if (compI->buildDate > maxSubDate) {
	        maxSubDate = compI->buildDate;
	    }
        }
        if (debugMode) { printf("}\n"); }
    }

    // see if the maximum build date of composite parts used by this assembly
    // is greater than the assembly's build date, and process the assembly if
    // this is found to be the case;  also, return the maximum build date for
    // this subassembly

    if (maxSubDate > assmH->buildDate) {
        assmH->DoNothing();
	count++;
    }
    return(maxSubDate);
}


//////////////////////////////////////////////////////////////////
//
// Top-Level Routine to do Query #6
//
//////////////////////////////////////////////////////////////////


int query6(ooHandle(ooDBObj)& databaseH)
{
    // scan modules, visiting all of the assemblies in each one, checking
    // to see if they use (either directly or indirectly) a more recently
    // dated composite part;  the visits are done by recursively traversing
    // the assemblies, checking them on the way back up (to avoid lots of
    // repeated work)

    ooItr(Module) modI;
    int count = 0;
    for (modI.scan(databaseH); modI.next(); ) {
	ooHandle(ComplexAssembly) assmH = modI->designRoot();
	int rslt = checkMaxCompDate(assmH, count);
    }
    return count;
}

